from neuralogic.dataset.base import BaseDataset, ConvertableDataset
from neuralogic.dataset.file import FileDataset
from neuralogic.dataset.logic import Dataset
from neuralogic.dataset.tensor import TensorDataset, Data
from neuralogic.dataset.csv import Mode, CSVFile, CSVDataset
from neuralogic.dataset.db import DBDataset, DBSource



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Activation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class SGConv(Module):
    r"""
    Simple Graph Convolutional layer from `"Simplifying Graph Convolutional Networks" <https://arxiv.org/abs/1902.07153>`_.
    Which can be expressed as:

    .. math::
        \mathbf{x}^{\prime}_i = act(\mathbf{W} \cdot {agg}_{j \in \mathcal{N}^k(i)}(\mathbf{x}_j))

    Where *act* is an activation function, *agg* aggregation function, *W* is a learnable parameter
    and :math:`\mathcal{N}^k(i)` denotes nodes that are *k* hops away from the node *i*.
    This equation is translated into the logic form as:

    .. code:: logtalk

        (R.<output_name>(V.I)[<W>] <= (
            R.<feature_name>(V.I<k>),
            R.<edge_name>(V.I<1>, V.I<0>), R.<edge_name>(V.I<2>, V.I<1>), ..., R.<edge_name>(V.I<k>, V.I<k-1>),
        )) | [<aggregation>, Activation.IDENTITY]

        R.<output_name> / 1 | [<activation>]

    Examples
    --------

    The whole computation of this module (parametrized as :code:`SGConv(2, 3, "h1", "h0", "_edge", 2)`) is as follows:

    .. code:: logtalk

        (R.h1(V.I0)[3, 2] <= (R.h0(V.I2), R._edge(V.I1, V.I0), R._edge(V.I2, V.I1))) | [Activation.IDENTITY, Aggregation.SUM]
        R.h1 / 1 | [Activation.IDENTITY]

    Module parametrized as :code:`SGConv(2, 3, "h1", "h0", "_edge", 1)` translates into:

    .. code:: logtalk

        (R.h1(V.I0)[3, 2] <= (R.h0(V.I1), R._edge(V.I1, V.I0))) | [Activation.IDENTITY, Aggregation.SUM]
        R.h1 / 1 | [Activation.IDENTITY]


    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    k : int
        Number of hops.
        Default: ``1``
    activation : Activation
        Activation function of the output.
        Default: ``Activation.IDENTITY``
    aggregation : Aggregation
        Aggregation function of nodes' neighbors.
        Default: ``Aggregation.SUM``

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        k: int = 1,
        activation: Activation = Activation.IDENTITY,
        aggregation: Aggregation = Aggregation.SUM,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.k = k

        self.aggregation = aggregation
        self.activation = activation

    def __call__(self):
        head = R.get(self.output_name)(V.I0)[self.out_channels, self.in_channels]
        metadata = Metadata(activation=Activation.IDENTITY, aggregation=self.aggregation, duplicit_grounding=True)
        edge = R.get(self.edge_name)
        feature = R.get(self.feature_name)

        return [
            (
                head
                <= (
                    feature(f"I{self.k}"),
                    *(edge(f"I{b}", f"I{a}") for a, b in zip(range(self.k), range(1, self.k + 1))),
                )
            )
            | metadata,
            R.get(self.output_name) / 1 | Metadata(activation=self.activation),
        ]



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Activation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module
from neuralogic.nn.module.general.rnn import RNNCell


class LSTMCell(Module):
    r"""

    Parameters
    ----------

    input_size : int
        Input feature size.
    hidden_size : int
        Output and hidden feature size.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input feature predicate name to get features from.
    hidden_input_name : str
        Predicate name to get hidden state from.
    cell_state_0_name : str
        Predicate name to get initial cell state from.
    arity : int
        Arity of the input and output predicate. Default: ``1``
    next_name : str
        Predicate name to get positive integer sequence from.
        Default: ``_next__positive``
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        output_name: str,
        input_name: str,
        hidden_input_name: str,
        cell_state_0_name: str,
        arity: int = 1,
        next_name: str = "_next__positive",
    ):
        self.input_size = input_size
        self.hidden_size = hidden_size

        self.output_name = output_name
        self.input_name = input_name
        self.hidden_input_name = hidden_input_name
        self.cell_state_0_name = cell_state_0_name

        self.arity = arity
        self.next_name = next_name

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity)]
        z_terms = [*terms, V.Z]
        t_terms = [*terms, V.T]

        i_name = f"{self.output_name}__i"
        f_name = f"{self.output_name}__f"
        g_name = f"{self.output_name}__n"
        o_name = f"{self.output_name}__o"

        c_left_name = f"{self.output_name}__left"
        c_right_name = f"{self.output_name}__right"
        c_name = f"{self.output_name}__c"
        c_tanh_name = f"{self.output_name}__ctanh"

        next_rel = R.get(self.next_name)(V.Z, V.T)

        cell_args = [
            self.input_size,
            self.hidden_size,
            i_name,
            self.input_name,
            self.hidden_input_name,
            Activation.SIGMOID,
            self.arity,
            self.next_name,
        ]

        i = RNNCell(*cell_args)

        cell_args[2] = f_name
        f = RNNCell(*cell_args)

        cell_args[2] = o_name
        o = RNNCell(*cell_args)

        cell_args[2] = g_name
        cell_args[-3] = Activation.TANH
        g = RNNCell(*cell_args)

        c_left = R.get(c_left_name)(t_terms) <= (R.get(f_name)(t_terms), R.get(c_name)(z_terms), next_rel)
        c_right = R.get(c_right_name)(t_terms) <= (R.get(i_name)(t_terms), R.get(g_name)(t_terms))
        c = R.get(c_name)(t_terms) <= (R.get(c_left_name)(t_terms), R.get(c_right_name)(t_terms))
        c_tanh = R.get(c_tanh_name)(t_terms) <= (R.get(c_name)(t_terms))
        h = R.get(self.output_name)(t_terms) <= (R.get(o_name)(t_terms), R.get(c_tanh_name)(t_terms))

        return [
            *i(),
            *f(),
            *o(),
            *g(),
            c_left | Metadata(activation="elementproduct-identity"),
            c_right | Metadata(activation="elementproduct-identity"),
            c_left.head.predicate | [Activation.IDENTITY],
            c_right.head.predicate | [Activation.IDENTITY],
            c | [Activation.IDENTITY],
            (R.get(c_name)([*terms, 0]) <= R.get(self.cell_state_0_name)(terms)) | [Activation.IDENTITY],
            c.head.predicate | [Activation.IDENTITY],
            c_tanh | [Activation.TANH],
            c_tanh.head.predicate | [Activation.IDENTITY],
            h | Metadata(activation="elementproduct-identity"),
            h.head.predicate | [Activation.IDENTITY],
        ]


class LSTM(Module):
    r"""
    One-layer Long Short-Term Memory (LSTM) RNN module which is computed as:

    .. math::

        i_t = \sigma(\mathbf{W}_{xi} \mathbf{x}_t + \mathbf{W}_{hi} \mathbf{h}_{t-1})

    .. math::

        f_t = \sigma(\mathbf{W}_{xf} \mathbf{x}_t + \mathbf{W}_{hf} \mathbf{h}_{t-1})

    .. math::

        o_t = \sigma(\mathbf{W}_{xo} \mathbf{x}_t + \mathbf{W}_{ho} \mathbf{h}_{t-1})


    .. math::

        g_t = \tanh(\mathbf{W}_{xg} \mathbf{x}_t + \mathbf{W}_{hg} \mathbf{h}_{t-1}) \\

    .. math::

        c_t = f_t \odot c_{t-1} + i_t \odot g_t

    .. math::

        h_t = o_t \odot \tanh(c_t)

    Parameters
    ----------

    input_size : int
        Input feature size.
    hidden_size : int
        Output and hidden feature size.
    sequence_length : int
        Sequence length.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input feature predicate name to get features from.
    hidden_0_name : str
        Predicate name to get initial hidden state from.
    cell_state_0_name : str
        Predicate name to get initial cell state from.
    arity : int
        Arity of the input and output predicate. Default: ``1``
    next_name : str
        Predicate name to get positive integer sequence from.
        Default: ``_next__positive``
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        sequence_length: int,
        output_name: str,
        input_name: str,
        hidden_0_name: str,
        cell_state_0_name: str,
        arity: int = 1,
        next_name: str = "_next__positive",
    ):
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.sequence_length = sequence_length

        self.output_name = output_name
        self.input_name = input_name
        self.hidden_0_name = hidden_0_name
        self.cell_state_0_name = cell_state_0_name

        self.arity = arity
        self.next_name = next_name

    def __call__(self):
        recursive_cell = LSTMCell(
            self.input_size,
            self.hidden_size,
            self.output_name,
            self.input_name,
            self.output_name,
            self.cell_state_0_name,
            self.arity,
            self.next_name,
        )

        next_relation = R.get(self.next_name)
        terms = [f"X{i}" for i in range(self.arity)]

        return [
            *[next_relation(i, i + 1) for i in range(0, self.sequence_length)],
            (R.get(self.output_name)([*terms, 0]) <= R.get(self.hidden_0_name)(terms)) | [Activation.IDENTITY],
            *recursive_cell(),
        ]



from neuralogic.core.enums import Backend


def get_neuralogic_layer(backend: Backend = Backend.JAVA):
    if backend == Backend.DYNET:
        from neuralogic.nn.dynet import NeuraLogic  # type: ignore

        return NeuraLogic
    if backend == Backend.JAVA:
        from neuralogic.nn.java import NeuraLogic  # type: ignore

        return NeuraLogic
    if backend == Backend.TORCH:
        from neuralogic.nn.torch import NeuraLogic  # type: ignore

        return NeuraLogic
    raise NotImplementedError


def get_evaluator(
    template,
    settings=None,
    backend: Backend = Backend.JAVA,
):
    from neuralogic.core.settings import Settings

    if settings is None:
        settings = Settings()

    if backend == Backend.DYNET:
        from neuralogic.nn.evaluator.dynet import DynetEvaluator

        return DynetEvaluator(template, settings)
    if backend == Backend.JAVA:
        from neuralogic.nn.evaluator.java import JavaEvaluator

        return JavaEvaluator(template, settings)
    if backend == Backend.TORCH:
        from neuralogic.nn.evaluator.torch import TorchEvaluator

        return TorchEvaluator(template, settings)



from typing import Callable


class Function:
    __slots__ = ("name",)

    def __init__(self, name: str):
        self.name: str = name.lower()

    def __str__(self):
        return self.name

    def pretty_str(self) -> str:
        return str(self).capitalize()

    def __call__(self, *args):
        if len(args) == 0 or args[0] is None:
            return self
        raise NotImplementedError


class Activation(Function):
    LUKASIEWICZ: "Activation" = None
    SIGMOID: "Activation" = None
    SIGNUM: "Activation" = None
    RELU: "Activation" = None
    LEAKY_RELU: "Activation" = None
    IDENTITY: "Activation" = None
    TANH: "Activation" = None
    EXP: "Activation" = None
    TRANSP: "Activation" = None
    NORM: "Activation" = None
    SQRT: "Activation" = None
    INVERSE: "Activation" = None
    REVERSE: "Activation" = None
    SOFTMAX: "Activation" = None
    SPARSEMAX: "Activation" = None

    __slots__ = ("nestable",)

    def __init__(self, name: str):
        super().__init__(name)
        self.nestable = True

    def nest(self, other: "Function") -> "Function":
        if isinstance(other, Aggregation):
            raise NotImplementedError(f"Cannot nest aggregation functions ({self} with {other})")

        if isinstance(other, Activation) and isinstance(self, Activation):
            raise NotImplementedError(f"Cannot nest activation functions ({self} with {other})")

        if not self.nestable:
            raise NotImplementedError(f"Cannot nest nested functions - {self} and {other}")

        function = Activation(f"{other.name}-{self.name}")
        function.nestable = False

        return function

    def __call__(self, *args):
        from neuralogic.core.constructs import relation

        if len(args) == 0 or args[0] is None:
            return self

        arg = args[0]

        if isinstance(arg, Callable) and not isinstance(arg, (ActivationAgg, relation.BaseRelation)):
            arg = arg()
        if isinstance(arg, relation.BaseRelation):
            return arg.attach_activation_function(self)
        if isinstance(arg, ActivationAgg):
            return self.nest(arg)
        raise NotImplementedError


Activation.LUKASIEWICZ = Activation("LUKASIEWICZ")
Activation.SIGMOID = Activation("SIGMOID")
Activation.SIGNUM = Activation("SIGNUM")
Activation.RELU = Activation("RELU")
Activation.LEAKY_RELU = Activation("LEAKYRELU")
Activation.IDENTITY = Activation("IDENTITY")
Activation.TANH = Activation("TANH")
Activation.EXP = Activation("EXP")
Activation.TRANSP = Activation("TRANSPOSE")
Activation.NORM = Activation("NORM")
Activation.SQRT = Activation("SQRT")
Activation.INVERSE = Activation("INVERSE")
Activation.REVERSE = Activation("REVERSE")
Activation.SOFTMAX = Activation("SOFTMAX")
Activation.SPARSEMAX = Activation("SPARSEMAX")


class ActivationAgg(Function):
    CROSSUM: "ActivationAgg" = None
    ELEMENTPRODUCT: "ActivationAgg" = None
    PRODUCT: "ActivationAgg" = None
    CONCAT: "ActivationAgg" = None
    MAX: "ActivationAgg" = None
    MIN: "ActivationAgg" = None

    def nest(self, other: "Function") -> "Function":
        if isinstance(other, Activation):
            return other.nest(self)

        if isinstance(other, ActivationAgg):
            raise NotImplementedError(f"Cannot nest act-aggregation functions ({self} with {other}")
        raise NotImplementedError

    def __str__(self):
        return f"{self.name}-identity"

    def __call__(self, *args):
        from neuralogic.core.constructs import relation

        if len(args) == 0 or args[0] is None:
            return self

        arg = args[0]

        if isinstance(arg, Callable) and not isinstance(arg, (Activation, relation.BaseRelation)):
            arg = arg()
        if isinstance(arg, relation.BaseRelation):
            return arg.attach_activation_function(self)
        if isinstance(arg, Activation):
            return self.nest(arg)
        raise NotImplementedError


ActivationAgg.CROSSUM = ActivationAgg("CROSSUM")
ActivationAgg.ELEMENTPRODUCT = ActivationAgg("ELEMENTPRODUCT")
ActivationAgg.PRODUCT = ActivationAgg("PRODUCT")
ActivationAgg.CONCAT = ActivationAgg("CONCAT")
ActivationAgg.MAX = ActivationAgg("MAX")
ActivationAgg.MIN = ActivationAgg("MIN")


class Aggregation(Function):
    MAX: "Aggregation" = None
    MIN: "Aggregation" = None
    AVG: "Aggregation" = None
    SUM: "Aggregation" = None


Aggregation.MAX = Aggregation("MAX")
Aggregation.MIN = Aggregation("MIN")
Aggregation.AVG = Aggregation("AVG")
Aggregation.SUM = Aggregation("SUM")



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Activation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module
from neuralogic.nn.module.general.rnn import RNNCell


class GRUCell(Module):
    r"""

    Parameters
    ----------

    input_size : int
        Input feature size.
    hidden_size : int
        Output and hidden feature size.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input feature predicate name to get features from.
    hidden_input_name : str
        Predicate name to get hidden state from.
    arity : int
        Arity of the input and output predicate. Default: ``1``
    next_name : str
        Predicate name to get positive integer sequence from.
        Default: ``_next__positive``
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        output_name: str,
        input_name: str,
        hidden_input_name: str,
        arity: int = 1,
        next_name: str = "_next__positive",
    ):
        self.input_size = input_size
        self.hidden_size = hidden_size

        self.output_name = output_name
        self.input_name = input_name
        self.hidden_input_name = hidden_input_name

        self.arity = arity
        self.next_name = next_name

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity)]
        p_terms = [*terms, V.Z]
        h_terms = [*terms, V.T]
        input_terms = [*terms, V.T]

        r_name = f"{self.output_name}__r"
        z_name = f"{self.output_name}__z"
        n_name = f"{self.output_name}__n"
        n_helper_name = f"{self.output_name}__n_helper"
        n_helper_weighted_name = f"{self.output_name}__n_helper_w"

        h_left_name = f"{self.output_name}__left"
        h_right_name = f"{self.output_name}__right"

        next_rel = R.get(self.next_name)(V.Z, V.T)

        r = RNNCell(
            self.input_size,
            self.hidden_size,
            r_name,
            self.input_name,
            self.hidden_input_name,
            Activation.SIGMOID,
            self.arity,
            self.next_name,
        )
        z = RNNCell(
            self.input_size,
            self.hidden_size,
            z_name,
            self.input_name,
            self.hidden_input_name,
            Activation.SIGMOID,
            self.arity,
            self.next_name,
        )

        h_weight = self.hidden_size, self.hidden_size
        n_helper = R.get(n_helper_name)(h_terms) <= (
            R.get(r_name)(h_terms),
            R.get(n_helper_weighted_name)(h_terms),
        )

        n_helper_weighted = R.get(n_helper_weighted_name)(h_terms) <= (
            R.get(self.hidden_input_name)(p_terms)[h_weight],
            next_rel,
        )

        i_weight = self.hidden_size, self.input_size
        n = R.get(n_name)(h_terms) <= (R.get(self.input_name)(input_terms)[i_weight], R.get(n_helper_name)(h_terms))

        h_left = R.get(h_left_name)(h_terms) <= (-R.get(z_name)(h_terms), R.get(n_name)(h_terms))
        h_right = R.get(h_right_name)(h_terms) <= (
            R.get(z_name)(h_terms),
            R.get(self.hidden_input_name)(p_terms),
            next_rel,
        )

        h = R.get(self.output_name)(h_terms) <= (R.get(h_left_name)(h_terms), R.get(h_right_name)(h_terms))

        return [
            *r(),
            *z(),
            n_helper | Metadata(activation="elementproduct-identity"),
            n_helper.head.predicate | [Activation.IDENTITY],
            n_helper_weighted | [Activation.IDENTITY],
            n_helper_weighted.head.predicate | [Activation.IDENTITY],
            n | [Activation.TANH],
            n.head.predicate | [Activation.IDENTITY],
            h_left | Metadata(activation="elementproduct-identity"),
            h_left.head.predicate | [Activation.IDENTITY],
            h_right | Metadata(activation="elementproduct-identity"),
            h_right.head.predicate | [Activation.IDENTITY],
            h | [Activation.IDENTITY],
            h.head.predicate | [Activation.IDENTITY],
        ]


class GRU(Module):
    r"""
    One-layer Gated Recurrent Unit (GRU) module which is computed as:

    .. math::

        r_t = \sigma(\mathbf{W}_{xr} \mathbf{x}_t + \mathbf{W}_{hr} \mathbf{h}_{t-1}) \\

    .. math::

        z_t = \sigma(\mathbf{W}_{xz} \mathbf{x}_t + \mathbf{W}_{hz} \mathbf{h}_{t-1}) \\

    .. math::

        n_t = \tanh(\mathbf{W}_{xn} \mathbf{x}_t + r_t \odot (\mathbf{W}_{hn} \mathbf{h}_{t-1})) \\

    .. math::

        h_t = (1 - z_t) \odot n_t + z_t \odot h_{t-1}

    where :math:`t \in (1, sequence\_length + 1)` is a time step.
    In the template, the :math:`t` is referred to as :code:`V.T`, and :math:`t - 1` is referred to as :code:`V.Z`.
    This module expresses the first equation as:

    .. code:: logtalk

        (R.<output_name>__r(<...terms>, V.T) <= (
            R.<input_name>(<...terms>, V.T)[<hidden_size>, <input_size>],
            R.<hidden_input_name>(<...terms>, V.Z)[<hidden_size>, <hidden_size>],
            R.<next_name>(V.Z, V.T),
        )) | [Activation.SIGMOID]

        R.<output_name>__r / <arity> + 1 | [Activation.IDENTITY]

    The second equation is expressed in the same way, except for a different head predicate name. The third equation is
    split into three rules. The first two computes the element-wise product -
    :math:`r_t * (\mathbf{W}_{hn} \mathbf{h}_{t-1})`.

    .. code:: logtalk

        (R.<output_name>__n_helper_weighted(<...terms>, V.T) <= (
            R.<hidden_input_name>(<...terms>, V.Z)[<hidden_size>, <hidden_size>], R.<next_name>(V.Z, V.T),
        )) | [Activation.IDENTITY],

        R.<output_name>__n_helper_weighted / (<arity> + 1) | [Activation.IDENTITY],

        (R.<output_name>__n_helper(<...terms>, V.T) <= (
            R.<output_name>__r(<..terms>, V.T), R.<>__n_helper_weighted(<...terms>, V.T)
        )) | Metadata(activation="elementproduct-identity"),

        R.<output_name>__n_helper / (<arity> + 1) | [Activation.IDENTITY],

    The third one computes the sum and applies the :math:`tanh` activation function.

    .. code:: logtalk

        (R.<output_name>__n(<...terms>, V.T) <= (
            R.<input_name>(<...terms>, V.T)[<hidden_size>, <input_size>],
            R.<output_name>__n_helper(<...terms>, V.T)
        )) | [Activation.TANH]
        R.<output_name>__n / (<arity> + 1) | [Activation.IDENTITY],

    The last equation is computed via three rules. The first two rules computes element-wise products. That is:

    .. code:: logtalk

        (R.<output_name>__left(<...terms>, V.T) <= (
            R.<output_name>__z(<...terms>, V.T), R.<output_name>__n(<...terms>, V.T)
        )) | Metadata(activation="elementproduct-identity")

        (R.<output_name>__right(<...terms>, V.T) <= (
            R.<output_name>__z(<...terms>, V.T), R.<hidden_input_name>(<...terms>, V.Z), R.<next_name>(V.Z, V.T),,
        )) | Metadata(activation="elementproduct-identity")

        R.<output_name>__left / <arity> + 1 | [Activation.IDENTITY]
        R.<output_name>__right / <arity> + 1 | [Activation.IDENTITY]

    The last output rule sums up the element-wise products.

    .. code:: logtalk

        (R.<output_name>(<...terms>, V.T) <= (
            R.<output_name>__left(<...terms>, V.T), R.<output_name>__right(<...terms>, V.T)
        )) | [Activation.IDENTITY]
        R.<output_name> / <arity> + 1 | [Activation.IDENTITY],

    Additionally, we define rules for the recursion purpose
    (the positive integer sequence :code:`R.<next_name>(V.Z, V.T)`) and the "stop condition", that is:

    .. code:: logtalk

        (R.<output_name>(<...terms>, 0) <= R.<hidden_0_name>(<...terms>)) | [Activation.IDENTITY]

    Parameters
    ----------

    input_size : int
        Input feature size.
    hidden_size : int
        Output and hidden feature size.
    sequence_length : int
        Sequence length.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input feature predicate name to get features from.
    hidden_0_name : str
        Predicate name to get initial hidden state from.
    arity : int
        Arity of the input and output predicate. Default: ``1``
    next_name : str
        Predicate name to get positive integer sequence from.
        Default: ``_next__positive``
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        sequence_length: int,
        output_name: str,
        input_name: str,
        hidden_0_name: str,
        arity: int = 1,
        next_name: str = "_next__positive",
    ):
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.sequence_length = sequence_length

        self.output_name = output_name
        self.input_name = input_name
        self.hidden_0_name = hidden_0_name

        self.arity = arity
        self.next_name = next_name

    def __call__(self):
        recursive_cell = GRUCell(
            self.input_size,
            self.hidden_size,
            self.output_name,
            self.input_name,
            self.output_name,
            self.arity,
            self.next_name,
        )

        next_relation = R.get(self.next_name)
        terms = [f"X{i}" for i in range(self.arity)]

        return [
            *[next_relation(i, i + 1) for i in range(0, self.sequence_length)],
            (R.get(self.output_name)([*terms, 0]) <= R.get(self.hidden_0_name)(terms)) | [Activation.IDENTITY],
            *recursive_cell(),
        ]



import json
from typing import Any, Dict, Optional

import numpy as np

from neuralogic.utils.visualize import draw_sample


class RawSample:
    __slots__ = "java_sample"

    def __init__(self, sample):
        self.java_sample = sample

    def draw(
        self,
        filename: Optional[str] = None,
        draw_ipython=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        return draw_sample(self, filename, draw_ipython, img_type, value_detail, graphviz_path, *args, **kwargs)


class Sample(RawSample):
    __slots__ = ("id", "target", "neurons", "output_neuron", "java_sample")

    def __init__(self, sample, java_sample):
        super().__init__(sample)
        serialized_sample = json.loads(str(sample.exportToJson()))

        self.id = serialized_sample["id"]
        self.target = json.loads(serialized_sample["target"])
        self.neurons = Sample.deserialize_network(serialized_sample["network"])
        self.output_neuron = self.neurons[-1].index
        self.java_sample = java_sample

    @staticmethod
    def deserialize_network(network):
        neurons = []

        for i, neuron in enumerate(network):
            neuron_object = Neuron(neuron, i)
            neurons.append(neuron_object)

        return neurons


class Neuron:
    def __init__(self, neuron: Dict[str, Any], index):
        self.index = index
        self.name = neuron["name"]
        self.weighted = neuron["weighted"]
        self.activation = neuron.get("activation", None)
        self.inputs = neuron["inputs"]
        self.weights = neuron.get("weights", None)
        self.offset = neuron["offset"]
        self.value = neuron.get("value", None)
        self.pooling = neuron["pooling"]
        self.hook_name = Neuron.parse_hook_name(self.name)

        if self.value:
            self.value = json.loads(self.value)

        if self.weights is not None:
            self.weights = list(self.weights)

        if self.inputs is not None:
            self.inputs = list(self.inputs)

    @staticmethod
    def parse_hook_name(name: str):
        name = name.split(" ")

        if len(name) == 3:
            return name[2]
        return None


class Weight:
    def __init__(self, weight):
        self.index: int = weight.index
        self.name = str(weight.name)
        self.dimensions = tuple(weight.value.size())
        self.value = json.loads(str(weight.value.toString()))
        self.fixed = weight.isFixed

        if not isinstance(self.value, list):
            self.value = self.value

        if not self.dimensions:
            self.dimensions = (1,)

        if self.fixed:
            self.value = np.array(self.value).reshape(self.dimensions)

    @staticmethod
    def get_unit_weight() -> "Weight":
        weight = Weight.__new__(Weight)
        weight.index = 0
        weight.name = "unit"
        weight.dimensions = (1,)
        weight.value = 1.0
        weight.fixed = True

        return weight


class BuiltDataset:
    """BuiltDataset represents an already built dataset - that is, a dataset that has been grounded and neuralized."""

    def __init__(self, samples):
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, item):
        return self.samples[item]



from neuralogic.nn.module.module import Module

from neuralogic.nn.module.gnn.appnp import APPNPConv
from neuralogic.nn.module.gnn.gatv2 import GATv2Conv
from neuralogic.nn.module.gnn.gcn import GCNConv
from neuralogic.nn.module.gnn.gin import GINConv
from neuralogic.nn.module.gnn.gsage import SAGEConv
from neuralogic.nn.module.gnn.res_gated import ResGatedGraphConv
from neuralogic.nn.module.gnn.rgcn import RGCNConv
from neuralogic.nn.module.gnn.sg import SGConv
from neuralogic.nn.module.gnn.tag import TAGConv

from neuralogic.nn.module.general.mlp import MLP
from neuralogic.nn.module.general.linear import Linear
from neuralogic.nn.module.general.rnn import RNN
from neuralogic.nn.module.general.gru import GRU
from neuralogic.nn.module.general.lstm import LSTM
from neuralogic.nn.module.general.rvnn import RvNN
from neuralogic.nn.module.general.pooling import Pooling, AvgPooling, SumPooling, MaxPooling

from neuralogic.nn.module.meta.meta import MetaConv
from neuralogic.nn.module.meta.magnn import MAGNNMean
from neuralogic.nn.module.meta.magnn import MAGNNLinear



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Activation, Aggregation, Function
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class RvNN(Module):
    r"""
    Recursive Neural Network (RvNN) module which is computed as:

    .. math::

         \mathbf{h}_i = act(agg_{j \in \mathcal{Ch(i)}}(\mathbf{W_{id(j)}} \mathbf{h}_j))

    Where :math:`act` is an activation function, :math:`agg` aggregation function and :math:`\mathbf{W}`'s
    are learnable parameters. :math:`\mathcal{Ch(i)}` represents the ordered list of children of node :math:`i`.
    The :math:`id(j)` function maps node :math:`j` to its index (position) in its parent's children list.

    Parameters
    ----------

    input_size : int
        Input feature size.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input feature predicate name to get leaf features from.
    parent_map_name : str
        Name of the predicate to get mapping from parent to children
    max_children : int
        Maximum number of children (specify which <max_children>-ary tree will be considered).
        Default: ``2``
    activation : Function
        Activation function of all layers.
        Default: ``Activation.TANH``
    aggregation : Function
        Aggregation function of a layer.
        Default: ``Activation.SUM``
    arity : int
        Arity of the input and output predicate (doesn't include the node id term). Default: ``1``
    """

    def __init__(
        self,
        input_size: int,
        output_name: str,
        input_name: str,
        parent_map_name: str,
        max_children: int = 2,
        activation: Function = Activation.TANH,
        aggregation: Function = Aggregation.SUM,
        arity: int = 1,
    ):
        self.input_size = input_size

        self.output_name = output_name
        self.input_name = input_name
        self.parent_map_name = parent_map_name
        self.max_children = max_children

        self.activation = activation
        self.aggregation = aggregation
        self.arity = arity

    def __call__(self):
        terms = [f"X{i}" for i in range(self.arity)]
        head_terms = [*terms, V.P]

        input_rel = R.get(self.input_name)
        output_rel = R.get(self.output_name)
        parent_map_rel = R.get(self.parent_map_name)
        metadata = Metadata(activation=self.activation, aggregation=self.aggregation)

        rules = [
            (output_rel(head_terms) <= (input_rel(head_terms), parent_map_rel(V.P))) | metadata,
            output_rel / len(head_terms) | [Activation.IDENTITY],
        ]

        body = []
        parent_terms = [V.P]

        for i in range(1, self.max_children + 1):
            term = f"C{i}"
            body.append(output_rel([*terms, term])[f"{self.output_name}__rvnn_{i}" : self.input_size, self.input_size])
            parent_terms.append(term)

            rules.append((output_rel(head_terms) <= (*body, parent_map_rel(*parent_terms))) | metadata)
        return rules



from typing import Iterable, Union

import numpy as np

from neuralogic.core.constructs.predicate import Predicate
from neuralogic.core.constructs import rule, factories
from neuralogic.core.constructs.function import Activation, ActivationAgg


class BaseRelation:
    __slots__ = "predicate", "function", "terms"

    def __init__(self, predicate: Predicate, terms=None, function: Union[Activation, ActivationAgg] = None):
        self.predicate = predicate
        self.function = function
        self.terms = terms

        if self.terms is None:
            self.terms = []
        elif not isinstance(self.terms, Iterable):
            self.terms = [self.terms]

    def __neg__(self) -> "BaseRelation":
        return self.__invert__()

    def __invert__(self) -> "BaseRelation":
        return self.attach_activation_function(Activation.REVERSE)

    @property
    def T(self) -> "BaseRelation":
        return self.attach_activation_function(Activation.TRANSP)

    def attach_activation_function(self, function: Union[Activation, ActivationAgg]):
        if self.function:
            function = function.nest(self.function)

        relation = self.__copy__()
        relation.function = function
        return relation

    def __truediv__(self, other):
        if not isinstance(other, int) or self.predicate.arity != 0 or other < 0:
            raise NotImplementedError

        name, hidden, special = self.predicate.name, self.predicate.hidden, self.predicate.special
        return factories.AtomFactory.Predicate.get_predicate(name, other, hidden, special)

    def __call__(self, *args) -> "BaseRelation":
        if self.terms:
            raise Exception

        if len(args) == 1 and isinstance(args[0], Iterable) and not isinstance(args[0], str):
            terms = list(args[0])
        else:
            terms = list(args)
        arity = len(terms)

        name, hidden, special = self.predicate.name, self.predicate.hidden, self.predicate.special
        predicate = factories.AtomFactory.Predicate.get_predicate(name, arity, hidden, special)

        return BaseRelation(predicate, terms, self.function)

    def __getitem__(self, item) -> "WeightedRelation":
        return WeightedRelation(item, self.predicate, False, self.terms, self.function)

    def __le__(self, other: Union[Iterable["BaseRelation"], "BaseRelation"]) -> rule.Rule:
        return rule.Rule(self, other)

    def to_str(self, end=False) -> str:
        end = "." if end else ""

        if self.terms:
            terms = ", ".join(str(term) for term in self.terms)
            name = f"{self.predicate.to_str()}({terms})"
        else:
            name = f"{self.predicate.to_str()}"

        if self.function:
            return f"{self.function}({name}){end}"
        return f"{name}{end}"

    def __str__(self) -> str:
        return self.to_str(True)

    def __copy__(self):
        atom = BaseRelation.__new__(BaseRelation)
        atom.function = self.function
        atom.terms = self.terms
        atom.predicate = self.predicate

        return atom


class WeightedRelation(BaseRelation):
    __slots__ = "weight", "weight_name", "is_fixed"

    def __init__(
        self, weight, predicate: Predicate, fixed=False, terms=None, function: Union[Activation, ActivationAgg] = None
    ):
        super().__init__(predicate, terms, function)

        self.weight = weight
        self.weight_name = None
        self.is_fixed = fixed

        if isinstance(weight, slice):
            self.weight_name = str(weight.start)
            self.weight = weight.stop
        elif isinstance(weight, tuple) and isinstance(weight[0], slice):
            self.weight_name = str(weight[0].start)
            self.weight = (weight[0].stop, *weight[1:])

        if isinstance(weight, np.ndarray):
            self.weight = weight.tolist()
        elif isinstance(weight, Iterable) and not isinstance(weight, tuple):
            self.weight = list(weight)

    def fixed(self) -> "WeightedRelation":
        if self.is_fixed:
            raise Exception(f"Weighted relation {self} is already fixed")
        return WeightedRelation(self.weight, self.predicate, True, self.terms, self.function)

    def to_str(self, end=False):
        if isinstance(self.weight, tuple):
            weight = f"{{{', '.join(str(w) for w in self.weight)}}}"
        else:
            weight = str(self.weight)
        if self.weight_name:
            weight = f"${self.weight_name}={weight}"

        if self.is_fixed:
            return f"<{weight}> {super().to_str(end)}"
        return f"{weight} {super().to_str(end)}"

    def __str__(self):
        return self.to_str(True)

    def __call__(self, *args) -> None:
        raise NotImplementedError(f"Cannot assign terms to weighted relation {self.predicate}")

    def __getitem__(self, item) -> None:
        raise NotImplementedError(f"Cannot assign weight to weighted relation {self.predicate}")

    def attach_activation_function(self, function: Union[Activation, ActivationAgg]):
        raise NotImplementedError(
            f"Cannot attach a function to weighted relation {self}. Attach the function before adding weights."
        )

    @property
    def T(self) -> "WeightedRelation":
        raise NotImplementedError(
            f"Cannot transpose weighted relation {self}. Apply the transposition before adding weights."
        )

    def __invert__(self) -> "WeightedRelation":
        raise NotImplementedError(f"Cannot negate weighted relation {self}. Apply the negation before adding weights.")

    def __neg__(self) -> "WeightedRelation":
        return self.__invert__()

    def __copy__(self):
        relation = WeightedRelation.__new__(WeightedRelation)

        relation.predicate = self.predicate
        relation.function = self.function
        relation.terms = self.terms
        relation.weight = self.weight
        relation.is_fixed = self.is_fixed

        return relation



from typing import Union

from neuralogic.core.constructs.relation import BaseRelation
from neuralogic.core.constructs.function import Activation, ActivationAgg, Function, Aggregation


# Activations


def lukasiewicz(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.LUKASIEWICZ(entity)


def sigmoid(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.SIGMOID(entity)


def signum(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.SIGNUM(entity)


def relu(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.RELU(entity)


def leaky_relu(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.LEAKY_RELU(entity)


def identity(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.IDENTITY(entity)


def tanh(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.TANH(entity)


def exp(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.EXP(entity)


def transp(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.TRANSP(entity)


def norm(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.NORM(entity)


def sqrt(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.SQRT(entity)


def inverse(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.INVERSE(entity)


def reverse(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.REVERSE(entity)


def softmax(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.SOFTMAX(entity)


def sparsemax(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return Activation.SPARSEMAX(entity)


# Activation-Aggregations


def crossum(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return ActivationAgg.CROSSUM(entity)


def elementproduct(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return ActivationAgg.ELEMENTPRODUCT(entity)


def product(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return ActivationAgg.PRODUCT(entity)


def concat(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return ActivationAgg.CONCAT(entity)


def max_act(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return ActivationAgg.MAX(entity)


def min_act(entity: Union[BaseRelation, Function] = None) -> Union[BaseRelation, Function]:
    return ActivationAgg.MIN(entity)


# Aggregations


def max(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.MAX(entity)


def min(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.MIN(entity)


def avg(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.AVG(entity)


def sum(entity: BaseRelation = None) -> Union[BaseRelation, Function]:
    return Aggregation.SUM(entity)



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Activation, Aggregation, Function
from neuralogic.core.constructs.factories import R
from neuralogic.nn.module.module import Module


class Pooling(Module):
    r"""
    Apply generic pooling over the input specified by the input_name and the input arity parameters.
    Can be expressed as:

    .. math::
        h = agg_{i_{0}, .., i_{n} \in N}(x_{(i_{0}, .., i_{n})})

    Where :math:`N` is a set of tuples of length :math:`n` (specified by the input arity parameter)
    that are valid arguments for the input predicate.

    For example, a classic pooling over graph nodes represented by relations of arity 1 (node id)
    would be calculated as:

    .. math::
        h = agg_{i \in N}(x_{(i)})

    Here :math:`N` refers to a set of all node ids. Lifting the restriction of the input arity via the input_arity
    parameter allows for pooling not only nodes but also edges (``input_arity=2``) and other objects (hyperedges etc.)

    Examples
    --------

    The whole computation of this module (parametrized as :code:`Pooling("h1", "h0", Aggregation.AVG)`) is as follows:

    .. code:: logtalk

        (R.h1 <= R.h0(V.X0)) | [Aggregation.AVG, Activation.IDENTITY]
        R.h1 / 0 | [Activation.IDENTITY]

    Module parametrized as :code:`Pooling("h1", "h0", Aggregation.MAX, 2)` translates into:

    .. code:: logtalk

        (R.h1 <= R.h0(V.X0, V.X1)) | [Aggregation.MAX, Activation.IDENTITY]
        R.h1 / 0 | [Activation.IDENTITY]

    Parameters
    ----------

    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input name.
    aggregation : Function
        Aggregation function.
    input_arity : int
        Arity of the input predicate ``input_name``. Default: ``1``
    """

    def __init__(
        self,
        output_name: str,
        input_name: str,
        aggregation: Function,
        input_arity: int = 1,
    ):
        self.output_name = output_name
        self.input_name = input_name
        self.input_arity = input_arity

        self.aggregation = aggregation

    def __call__(self):
        metadata = Metadata(activation=Activation.IDENTITY, aggregation=self.aggregation)

        return [
            (R.get(self.output_name) <= R.get(self.input_name)(f"X{i}" for i in range(self.input_arity))) | metadata,
            R.get(self.output_name) / 0 | [Activation.IDENTITY],
        ]


class MaxPooling(Pooling):
    r"""
    Apply max pooling over the input specified by the input_name and the input arity parameters.
    Can be expressed as:

    .. math::
        h = max_{i_{0}, .., i_{n} \in N}(x_{(i_{0}, .., i_{n})})

    Where :math:`N` is a set of tuples of length :math:`n` (specified by the input arity parameter)
    that are valid arguments for the input predicate.

    This module extends the generic pooling :class:`~neuralogic.nn.module.pooling.Pooling`.

    Examples
    --------

    The whole computation of this module (parametrized as :code:`MaxPooling("h1", "h0")`) is as follows:

    .. code:: logtalk

        (R.h1 <= R.h0(V.X0)) | [Aggregation.MAX, Activation.IDENTITY]
        R.h1 / 0 | [Activation.IDENTITY]

    Parameters
    ----------

    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input name.
    input_arity : int
        Arity of the input predicate ``input_name``. Default: ``1``
    """

    def __init__(self, output_name: str, input_name: str, input_arity: int = 1):
        super().__init__(output_name, input_name, Aggregation.MAX, input_arity)


class AvgPooling(Pooling):
    r"""
    Apply average pooling over the input specified by the input_name and the input arity parameters.
    Can be expressed as:

    .. math::
        h = \frac{1}{|N|}\sum_{i_{0}, .., i_{n} \in N} x_{(i_{0}, .., i_{n})}

    Where :math:`N` is a set of tuples of length :math:`n` (specified by the input arity parameter)
    that are valid arguments for the input predicate.

    This module extends the generic pooling :class:`~neuralogic.nn.module.pooling.Pooling`.

    Examples
    --------

    The whole computation of this module (parametrized as :code:`AvgPooling("h1", "h0")`) is as follows:

    .. code:: logtalk

        (R.h1 <= R.h0(V.X0)) | [Aggregation.AVG, Activation.IDENTITY]
        R.h1 / 0 | [Activation.IDENTITY]

    Parameters
    ----------

    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input name.
    input_arity : int
        Arity of the input predicate ``input_name``. Default: ``1``
    """

    def __init__(self, output_name: str, input_name: str, input_arity: int = 1):
        super().__init__(output_name, input_name, Aggregation.AVG, input_arity)


class SumPooling(Pooling):
    r"""
    Apply sum pooling over the input specified by the input_name and the input arity parameters.
    Can be expressed as:

    .. math::
        h = \sum_{i_{0}, .., i_{n} \in N} x_{(i_{0}, .., i_{n})}

    Where :math:`N` is a set of tuples of length :math:`n` (specified by the input arity parameter)
    that are valid arguments for the input predicate.

    This module extends the generic pooling :class:`~neuralogic.nn.module.pooling.Pooling`.

    Examples
    --------

    The whole computation of this module (parametrized as :code:`SumPooling("h1", "h0")`) is as follows:

    .. code:: logtalk

        (R.h1 <= R.h0(V.X0)) | [Aggregation.SUM, Activation.IDENTITY]
        R.h1 / 0 | [Activation.IDENTITY]

    Parameters
    ----------

    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input name.
    input_arity : int
        Arity of the input predicate ``input_name``. Default: ``1``
    """

    def __init__(self, output_name: str, input_name: str, input_arity: int = 1):
        super().__init__(output_name, input_name, Aggregation.SUM, input_arity)



import os
from enum import Enum

import jpype


_is_logging_initialized = False
_loggers_buffer = []


class TextIOWrapper:
    def __init__(self, wrapped_text_io):
        self.wrapped_text_io = wrapped_text_io

    def write(self, string):
        self.wrapped_text_io.write(str(string))


class Level(Enum):
    """Logging level"""

    OFF = "OFF"
    SEVERE = "SEVERE"
    WARNING = "WARNING"
    INFO = "INFO"
    CONFIG = "CONFIG"
    FINE = "FINE"
    FINER = "FINER"
    FINEST = "FINEST"
    ALL = "ALL"


class Formatter(Enum):
    """Logged information formatters"""

    COLOR = "color"
    NORMAL = "normal"


def _init_logging():
    global _is_logging_initialized

    jpype.java.lang.System.setOut(jpype.java.io.PrintStream(os.devnull))
    jpype.java.lang.System.setErr(jpype.java.io.PrintStream(os.devnull))

    jpype.JClass("cz.cvut.fel.ida.logging.Logging").initLogging(jpype.JClass("cz.cvut.fel.ida.setup.Settings")())
    _is_logging_initialized = True

    for handler_settings in _loggers_buffer:
        add_handler(*handler_settings)
    _loggers_buffer.clear()


def add_handler(output, level: Level = Level.FINER, formatter: Formatter = Formatter.COLOR):
    """
    Add logger handler for an insight into the java backend

    :param output: File-like object (has write(text: str) method)
    :param level: The logging level
    :param formatter: The log formatter
    """
    if not _is_logging_initialized:
        _loggers_buffer.append((output, level, formatter))
        return

    root_logger = jpype.java.util.logging.Logger.getLogger("")
    wrapped = TextIOWrapper(output)

    java_io_wrapper = jpype.JProxy("cz.cvut.fel.ida.utils.python.PythonOutputStream.TextIOWrapper", inst=wrapped)
    java_output_stream = jpype.JClass("cz.cvut.fel.ida.utils.python.PythonOutputStream")(java_io_wrapper)

    if formatter == Formatter.COLOR:
        java_formatter = jpype.JClass("cz.cvut.fel.ida.logging.ColoredFormatter")()
    elif formatter == Formatter.NORMAL:
        java_formatter = jpype.JClass("cz.cvut.fel.ida.logging.NormalFormatter")()
    else:
        raise NotImplementedError(f"Unknown formatter {formatter}")

    print_stream = jpype.java.io.PrintStream(java_output_stream)
    stream_handler = jpype.JClass("cz.cvut.fel.ida.logging.FlushStreamHandler")(print_stream, java_formatter)
    stream_handler.setLevel(getattr(jpype.java.util.logging.Level, str(level.value)))

    root_logger.addHandler(stream_handler)


def clear_handlers():
    """Clear all handlers"""
    if not _is_logging_initialized:
        _loggers_buffer.clear()
        return

    root_logger = jpype.java.util.logging.Logger.getLogger("")

    for handler in root_logger.getHandlers():
        root_logger.removeHandler(handler)



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Activation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class ResGatedGraphConv(Module):
    r"""
    Residual Gated Graph Convolutional layer from `"Residual Gated Graph ConvNets" <https://arxiv.org/abs/1711.07553>`_.
    Which can be expressed as:

    .. math::

        \mathbf{x}^{\prime}_i = act(\mathbf{W}_1 \mathbf{x}_i +
         {agg}_{j \in \mathcal{N}(i)}(\eta_{i,j} \odot \mathbf{W}_2 \mathbf{x}_j))

    .. math::

        \mathbf{\eta}_{i,j} = gating\_act(\mathbf{W}_3 \mathbf{x}_i + \mathbf{W}_4 \mathbf{x}_j)

    Where *act* is an activation function, *agg* aggregation function, *gating_act* is a gating activation function and
    :math:`W_n` are learnable parameters. This equation is translated into the logic form as:

    .. code:: logtalk

        (R.<output_name>__gate(V.I, V.J) <= (R.<feature_name>(V.I)[<W>], R.<feature_name>(V.J)[<W>])) | [Activation.IDENTITY]
        R.<output_name>__gate / 2 | [<activation>]

        (R.<output_name>(V.I) <= R.<feature_name>(V.I)[<W>]) | [Activation.IDENTITY]
        (R.<output_name>(V.I) <= (
            R.<output_name>__gate(V.I, V.J), R.<feature_name>(V.J)[<W>], R.<edge_name>(V.J, V.I))
        ) | Metadata(activation="elementproduct-identity", aggregation=<aggregation>)

        R.<output_name> / 1 | [<activation>]

    Examples
    --------

    The whole computation of this module (parametrized as :code:`ResGatedGraphConv(1, 2, "h1", "h0", "_edge")`)
    is as follows:

    .. code:: logtalk

        metadata = Metadata(activation="elementproduct-identity", aggregation=Aggregation.SUM)

        (R.h1__gate(V.I, V.J) <= (R.h0(V.I)[2, 1], R.h0(V.J)[2, 1])) | [Activation.IDENTITY]
        R.h1__gate / 2 | [Activation.SIGMOID]

        (R.h1(V.I) <= R.h0(V.I)[2, 1]) | [Activation.IDENTITY]
        (R.h1(V.I) <= (R.h1__gate(V.I, V.J), R.h0(V.J)[2, 1], R._edge(V.J, V.I))) | metadata
        R.h1 / 1 | [Activation.IDENTITY]

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    gating_activation : Activation
        Gating activation function.
        Default: ``Activation.SIGMOID``
    activation : Activation
        Activation function of the output.
        Default: ``Activation.IDENTITY``
    aggregation : Aggregation
        Aggregation function of nodes' neighbors.
        Default: ``Aggregation.SUM``

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        gating_activation: Activation = Activation.SIGMOID,
        activation: Activation = Activation.IDENTITY,
        aggregation: Aggregation = Aggregation.SUM,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.aggregation = aggregation
        self.activation = activation
        self.gating_activation = gating_activation

    def __call__(self):
        head = R.get(self.output_name)(V.I)
        feature = R.get(self.feature_name)
        gate = R.get(f"{self.output_name}__gate")

        prod_metadata = Metadata(activation="elementproduct-identity", aggregation=self.aggregation)
        w = self.out_channels, self.in_channels

        return [
            (gate(V.I, V.J) <= (feature(V.I)[w], feature(V.J)[w])) | [Activation.IDENTITY],
            gate / 2 | Metadata(activation=self.gating_activation),
            (head <= feature(V.I)[w]) | [Activation.IDENTITY],
            (head <= (gate(V.I, V.J), feature(V.J)[w], R.get(self.edge_name)(V.J, V.I))) | prod_metadata,
            R.get(self.output_name) / 1 | Metadata(activation=self.activation),
        ]



from typing import Optional, Dict, Union

import dynet as dy
from neuralogic.nn.loss import ErrorFunctionNames

from neuralogic.nn.base import AbstractEvaluator

from neuralogic.core import Template, BuiltDataset
from neuralogic.dataset import BaseDataset
from neuralogic.core.settings import Settings
from neuralogic.core.enums import Backend, Optimizer


class DynetEvaluator(AbstractEvaluator):
    trainers = {
        Optimizer.SGD: lambda param, rate: dy.SimpleSGDTrainer(param, learning_rate=rate),
        Optimizer.ADAM: lambda param, rate: dy.AdamTrainer(param, alpha=rate),
    }

    error_functions = {
        str(ErrorFunctionNames.MSE): lambda out, target: dy.squared_distance(out, target),
        # ErrorFunction.ABS_DIFF: lambda out, target: dy.abs(out - target),
        # ErrorFunction.CROSSENTROPY: lambda out, target: pass
    }

    def __init__(
        self,
        template: Template,
        settings: Settings,
    ):
        super().__init__(Backend.DYNET, template, settings)

    def train(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        epochs = self.settings.epochs
        error_function = str(self.settings.error_function)
        optimizer = Optimizer[str(self.settings.optimizer)]

        if optimizer not in DynetEvaluator.trainers:
            raise NotImplementedError
        if error_function not in DynetEvaluator.error_functions:
            raise NotImplementedError

        trainer = DynetEvaluator.trainers[optimizer](self.neuralogic_model.model, self.settings.learning_rate)
        error_function = DynetEvaluator.error_functions[error_function]

        def _train():
            for _ in range(epochs):
                seen_instances = 0
                total_loss = 0

                dy.renew_cg(immediate_compute=False, check_validity=False)

                for sample in dataset.samples:
                    if isinstance(sample.target, list):
                        label = dy.inputTensor(sample.target)
                    else:
                        label = dy.scalarInput(sample.target)
                    graph_output = self.neuralogic_model(sample)

                    loss = error_function(graph_output, label)

                    total_loss += loss.value()
                    loss.backward()
                    trainer.update()
                    seen_instances += 1

                yield total_loss, seen_instances

        if generator:
            return _train()

        stats = 0, 0
        for stats in _train():
            pass
        return stats

    def test(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        def _test():
            for sample in dataset.samples:
                dy.renew_cg(immediate_compute=False, check_validity=False)

                graph_output = self.neuralogic_model(sample)
                yield graph_output.value()

        if generator:
            return _test()
        return list(_test())

    def state_dict(self) -> Dict:
        return self.neuralogic_model.state_dict()

    def load_state_dict(self, state_dict: Dict):
        self.neuralogic_model.load_state_dict(state_dict)



from typing import List

import jpype

from neuralogic import is_initialized, initialize
from neuralogic.core.builder.components import Weight, Sample, RawSample
from neuralogic.core.enums import Backend
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.sources import Sources


def stream_to_list(stream) -> List:
    return list(stream.collect(jpype.JClass("java.util.stream.Collectors").toList()))


class Builder:
    def __init__(self, settings: SettingsProxy):
        if not is_initialized():
            initialize()

        self.settings = settings
        self.example_builder = Builder.get_builders(settings)
        self.builder = Builder.get_builders(settings)

        self.neural_model = jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.training.NeuralModel")
        self.collectors = jpype.JClass("java.util.stream.Collectors")

    def build_template_from_file(self, settings: SettingsProxy, filename: str):
        args = [
            "-t",
            filename,
            "-q",
            filename,
        ]

        sources = Sources.from_args(args, settings)
        template = self.builder.buildTemplate(sources.sources)

        return template

    def from_sources(self, parsed_template, sources: Sources, backend: Backend):
        source_pipeline = self.example_builder.buildPipeline(parsed_template, sources.sources)
        source_pipeline.execute(None if sources is None else sources.sources)
        java_model = source_pipeline.get()

        logic_samples = java_model.s
        logic_samples = logic_samples.collect(self.collectors.toList())

        if backend == Backend.JAVA:
            return [RawSample(sample) for sample in logic_samples]
        return Builder.build(logic_samples)

    def from_logic_samples(self, parsed_template, logic_samples, backend: Backend):
        source_pipeline = self.example_builder.buildPipeline(parsed_template, logic_samples)
        source_pipeline.execute(None)
        java_model = source_pipeline.get()

        logic_samples = java_model.s
        logic_samples = logic_samples.collect(self.collectors.toList())

        if backend == Backend.JAVA:
            return [RawSample(sample) for sample in logic_samples]

        return Builder.build(logic_samples)

    def build_model(self, parsed_template, backend: Backend, settings: SettingsProxy):
        neural_model = self.neural_model(parsed_template.getAllWeights(), settings.settings)

        if backend == Backend.JAVA:
            return neural_model

        dummy_weight = Weight.get_unit_weight()
        weights: List = [dummy_weight] * len(parsed_template.getAllWeights())

        for x in parsed_template.getAllWeights():
            weight = Weight(x)

            if weight.index >= len(weights):
                weights.extend([dummy_weight] * (weight.index - len(weights) + 1))
            weights[weight.index] = weight
        return weights

    @staticmethod
    def get_builders(settings: SettingsProxy):
        builder = jpype.JClass("cz.cvut.fel.ida.pipelines.building.PythonBuilder")(settings.settings)

        return builder

    @staticmethod
    def build(samples):
        serializer = jpype.JClass("cz.cvut.fel.ida.neural.networks.structure.export.NeuralSerializer")()
        super_detailed_format = jpype.JClass("cz.cvut.fel.ida.setup.Settings").superDetailedNumberFormat
        serializer.numberFormat = super_detailed_format

        return [Sample(serializer.serialize(sample), sample) for sample in samples]



from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Activation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class SAGEConv(Module):
    r"""
    GraphSAGE layer from `"Inductive Representation Learning on Large Graphs" <https://arxiv.org/abs/1706.02216>`_.
    Which can be expressed as:

    .. math::
        \mathbf{x}^{\prime}_i = act(\mathbf{W}_1 \mathbf{x}_i + \mathbf{W}_2 \cdot
         {agg}_{j \in \mathcal{N}(i)}(\mathbf{x}_j)))

    Where *act* is an activation function, *agg* aggregation function and *W*'s are learnable parameters.
    This equation is translated into the logic form as:

    .. code:: logtalk

         (R.<output_name>(V.I)[<W1>] <= (R.<feature_name>(V.J), R.<edge_name>(V.J, V.I))) | [<aggregation>, Activation.IDENTITY]
         (R.<output_name>(V.I)[<W2>] <= R.<feature_name>(V.I)) | [Activation.IDENTITY]
         R.<output_name> / 1 | [<activation>]

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : str
        Edge predicate name to use for neighborhood relations.
    activation : Activation
        Activation function of the output.
        Default: ``Activation.IDENTITY``
    aggregation : Aggregation
        Aggregation function of nodes' neighbors.
        Default: ``Aggregation.AVG``

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: str,
        activation: Activation = Activation.IDENTITY,
        aggregation: Aggregation = Aggregation.AVG,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.aggregation = aggregation
        self.activation = activation

    def __call__(self):
        head = R.get(self.output_name)(V.I)[self.out_channels, self.in_channels]
        metadata = Metadata(activation=Activation.IDENTITY, aggregation=self.aggregation)

        return [
            (head <= (R.get(self.feature_name)(V.J), R.get(self.edge_name)(V.J, V.I))) | metadata,
            (head <= R.get(self.feature_name)(V.I)) | metadata,
            R.get(self.output_name) / 1 | Metadata(activation=self.activation),
        ]



import enum
from pathlib import Path
from typing import Optional, List, Union, TextIO, Callable

from neuralogic.core.constructs.factories import R
from neuralogic.core.constructs.relation import BaseRelation, WeightedRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.dataset import Dataset
from neuralogic.dataset.base import ConvertableDataset

DatasetEntries = Union[BaseRelation, WeightedRelation, Rule]


class Mode(enum.Enum):
    ONE_EXAMPLE = "one"
    EXAMPLE_PER_SOURCE = "example_per_source"
    ZIP = "zip"


class CSVFile:
    __slots__ = (
        "relation_name",
        "csv_source",
        "sep",
        "value_column",
        "default_value",
        "value_mapper",
        "term_columns",
        "header",
        "skip_rows",
        "n_rows",
        "replace_empty_column",
    )

    def __init__(
        self,
        relation_name: str,
        csv_source: Union[TextIO, Path],
        sep=",",
        value_column: Optional[Union[str, int]] = None,
        default_value: Optional[Union[float, int]] = None,
        value_mapper: Optional[Callable] = None,
        term_columns: Optional[List[Union[str, int]]] = None,
        header: bool = False,
        skip_rows: int = 0,
        n_rows: Optional[int] = None,
        replace_empty_column: Union[str, float, int] = 0,
    ):
        self.relation_name = relation_name
        self.csv_source = csv_source
        self.sep = sep
        self.value_column = value_column
        self.default_value = default_value
        self.value_mapper = value_mapper
        self.term_columns = term_columns
        self.header = header
        self.skip_rows = skip_rows
        self.n_rows = n_rows
        self.replace_empty_column = replace_empty_column

    @staticmethod
    def _find_index_in_header(header, value) -> int:
        for index, header_value in enumerate(header):
            if value == header_value:
                return index
        raise ValueError(f"Value {value} not found in the header {header}")

    def _get_column_indices(self, header) -> Optional[List[int]]:
        if self.term_columns is None:
            return None
        new_columns = []

        for col_value in self.term_columns:
            new_columns.append(CSVFile._find_index_in_header(header, col_value))
        return new_columns

    def _to_logic(self, fp: TextIO) -> List[DatasetEntries]:
        example = []

        use_columns = self.term_columns
        value_column = self.value_column
        default_value = self.default_value
        value_mapper = self.value_mapper
        relation = R.get(self.relation_name)
        replace_empty = self.replace_empty_column
        read_lines = 0

        if self.header:
            header = fp.readline()
            if not header:
                return example
            header = header.strip().split(self.sep)

            value_column = None if value_column is None else CSVFile._find_index_in_header(header, value_column)
            use_columns = self._get_column_indices(header)

        for _ in range(self.skip_rows):
            fp.readline()

        while True:
            line = fp.readline()

            if not line or not line.strip():
                break

            terms = line.strip().split(self.sep)
            if use_columns is None:
                line_relation = relation([(term.strip() if len(term.strip()) else replace_empty) for term in terms])
            else:
                line_relation = relation(
                    [(terms[i].strip() if len(terms[i].strip()) else replace_empty) for i in use_columns]
                )

            if value_column is None:
                if default_value is not None:
                    line_relation = line_relation[float(default_value)]
            else:
                value = terms[value_column].strip()
                if not len(value):
                    value = default_value if default_value is not None else replace_empty
                if value_mapper is None:
                    line_relation = line_relation[float(value)]
                else:
                    line_relation = line_relation[value_mapper(value)]

            example.append(line_relation)
            read_lines += 1
            if read_lines == self.n_rows:
                break
        return example

    def to_logic_form(self) -> List[DatasetEntries]:
        if isinstance(self.csv_source, (str, Path)):
            with open(self.csv_source, "r") as fp:
                return self._to_logic(fp)
        return self._to_logic(self.csv_source)


class CSVDataset(ConvertableDataset):
    def __init__(
        self,
        csv_files: Union[List[CSVFile], CSVFile],
        csv_queries: Optional[CSVFile] = None,
        mode: Mode = Mode.ONE_EXAMPLE,
    ):
        self.csv_queries = csv_queries
        self.csv_files = [csv_files] if isinstance(csv_files, CSVFile) else csv_files
        self.mode = mode

    def add_csv_file(self, file: CSVFile):
        self.csv_files.append(file)

    def set_query_csv_file(self, file: CSVFile):
        self.csv_queries = file

    def to_dataset(self) -> Dataset:
        examples = []
        queries = self.csv_queries.to_logic_form() if self.csv_queries else []
        dataset = Dataset(examples, queries)

        if self.mode == Mode.ONE_EXAMPLE:
            example = []

            for source in self.csv_files:
                example.extend(source.to_logic_form())
            examples.append(example)
        elif self.mode == Mode.ZIP:
            logic_examples = [source.to_logic_form() for source in self.csv_files]

            for zipped_example in zip(*logic_examples):
                examples.append(zipped_example)
        elif self.mode == Mode.EXAMPLE_PER_SOURCE:
            for source in self.csv_files:
                examples.append(source.to_logic_form())
        return dataset



from neuralogic.core.constructs.factories import Var, Constant, Relation, V, C, R
from neuralogic.core.template import Template
from neuralogic.core.builder import BuiltDataset
from neuralogic.core.constructs.java_objects import JavaFactory
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.settings import Settings, SettingsProxy
from neuralogic.core.enums import Optimizer, Backend
from neuralogic.core.constructs.function import Activation, Aggregation, ActivationAgg



from typing import List, Union

from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Activation, Function
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class MLP(Module):
    r"""

    Parameters
    ----------

    units : List[int]
        List of layer sizes.
    output_name : str
        Output (head) predicate name of the module.
    input_name : str
        Input name.
    activation : Union[Function, List[Function]]
        Activation function of all layers or list of activations for each layer.
        Default: ``Activation.RELU``
    """

    def __init__(
        self,
        units: List[int],
        output_name: str,
        input_name: str,
        activation: Union[Function, List[Function]] = Activation.RELU,
    ):
        self.output_name = output_name
        self.input_name = input_name

        self.units = units
        self.activation: Union[Function, List[Function]] = activation

    def __call__(self):
        layers = []

        prev_layer = R.get(self.input_name)

        if isinstance(self.activation, list):
            metadata = None
        else:
            metadata = Metadata(activation=self.activation)

        for index, (in_channels, out_channels) in enumerate(zip(self.units[:-1], self.units[1:])):
            out_layer = R.get(f"{self.output_name}__{index}")

            layers.append(out_layer(V.I)[out_channels, in_channels] <= prev_layer(V.I))
            act_layer = out_layer / 1 | (Metadata(activation=self.activation[index]) if metadata is None else metadata)

            layers.append(act_layer)
            prev_layer = out_layer

        in_channels, out_channels = self.units[-2], self.units[-1]
        out = R.get(self.output_name)

        layers.append(out(V.I)[out_channels, in_channels] <= prev_layer(V.I))
        layers.append(out / 1 | (Metadata(activation=self.activation[-1]) if metadata is None else metadata))

        return layers
